from ..downloader import OpenMLDownloader
from ..downloader import RDataDownloader
from ..downloader import KaggleDownloader
from ..downloader import UCIMLDownloader
from .base_dataset import Dataset
import pandas as pd
import os
import pyreadr
import numpy as np

from config import DATA_DIR, DATA_DOWNLOAD_DIR


def count_show_unique(data):
    for col in data.columns:
        unique_values = data[col].unique().tolist()
        data_type = data[col].dtype
        num_unique = data[col].nunique()
        if num_unique < 20:
            print(f"{col} ({data_type}) => {num_unique} values: {unique_values}")
        else:
            print(f"{col} ({data_type}) => {num_unique} values")

    
# todo: add codes for categorical features

class KidneyDiseaseDataset(Dataset):

    def __init__(self):
        super().__init__()
        self.name = 'kidney_disease'
        self.data_dir = os.path.join(DATA_DIR, self.name)

    def load(self):
        
        # download data
        raw_data = self.download_data()

        # specify meta data
        self.sensitive_features = ['age']
        self.drop_features = ['id']
        self.ordinal_features = ['al', 'su']
        self.binary_features = ['rbc', 'pc', 'pcc', 'ba', 'htn', 'dm', 'cad', 'appet', 'pe', 'ane', 'classification']
        self.multiclass_features = []
        self.target_features = ['classification']
        self.numerical_features = [
            col for col in raw_data.columns 
            if col not in self.binary_features + self.multiclass_features + self.ordinal_features + self.drop_features
        ]

        self.raw_data = raw_data
        self.raw_data = self.custom_raw_data_preprocessing(self.raw_data)
        # basic processing
        self.raw_data = self.basic_processing(self.raw_data)

    def handle_missing_data(self, data: pd.DataFrame):
        return data.dropna()
    
    def download_data(self):
        downloader = KaggleDownloader(
            dataset_name='mansoordaku/ckdisease',
            file_names = ['kidney_disease.csv'],
        )

        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        raw_data = pd.read_csv(os.path.join(self.data_dir, DATA_DOWNLOAD_DIR, 'kidney_disease.csv'))

        return raw_data
    
    def custom_raw_data_preprocessing(self, raw_data: pd.DataFrame):

        # categories error correction
        corretion_cols = ['dm', 'cad']
        for col in corretion_cols:
            raw_data[col] = raw_data[col].replace({' yes':  'yes', '\tyes': 'yes', '\tno': 'no'})
        
        raw_data['classification'] = raw_data['classification'].replace({'ckd\t': 'ckd'})

        # numerical feature error correction
        corretion_cols = ['pcv', 'wc', 'rc']
        for col in corretion_cols:
            raw_data[col] = raw_data[col].replace('\t?', np.nan)
            raw_data[col] = raw_data[col].replace('\t[0-9]+', lambda x: x.group(0).replace('\t', ''))
            raw_data[col] = pd.to_numeric(raw_data[col], errors='coerce')

        return raw_data
    


# class BreastCancerLjubljanaDataset(Dataset):

#     def __init__(self):
#         super().__init__()

#     def load(self):
#         pass

#     def handle_missing_data(self, data: pd.DataFrame):
#         pass

#     def custom_download(self):
#         import requests
#         import zipfile
#         import tempfile
#         import io
#         import os
#         import shutil
#         url = 'https://archive.ics.uci.edu/static/public/14/breast+cancer.zip'
#         response = requests.get(url)
#         zip_content = io.BytesIO(response.content)
#         temp_dir = tempfile.mkdtemp()
#         with zipfile.ZipFile(zip_content) as zip_ref:
#             zip_ref.extractall(temp_dir)
        
#         file_name = 'breast-cancer.data'
#         file_path = os.path.join(temp_dir, file_name)
#         if file_name.endswith('.data'):
#             data = pd.read_csv(file_path, header=None)

#         # remove temp_dir
#         shutil.rmtree(temp_dir)

#         return data
    
class ParkinsonsTelemonitoringDataset(Dataset):

    def __init__(self):
        super().__init__()
        self.name = 'parkinsons_telemonitoring'
        self.data_dir = os.path.join(DATA_DIR, self.name)

    def load_raw_data(self):
        raw_data = self.download_data()
        return raw_data

    def load(self):

        # download data
        raw_data = self.download_data()

        # specify meta data
        self.raw_data = raw_data
        self.sensitive_features = ['age', 'sex']
        self.drop_features = []
        self.ordinal_features = []
        self.binary_features = ['sex']
        self.multiclass_features = []
        self.target_features = ['total_UPDRS', 'motor_UPDRS']
        self.numerical_features = [
            col for col in raw_data.columns 
            if col not in self.binary_features + self.multiclass_features + self.ordinal_features + self.drop_features
        ]

        # basic processing
        self.raw_data = self.basic_processing(raw_data)

    def handle_missing_data(self, data: pd.DataFrame):
        return data.dropna()

    def download_data(self):
        downloader = UCIMLDownloader(
            url = 'https://archive.ics.uci.edu/static/public/189/parkinsons+telemonitoring.zip'
        )

        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        raw_data = pd.read_csv(os.path.join(self.data_dir, DATA_DOWNLOAD_DIR, 'parkinsons_updrs.data'), index_col = 0)
        raw_data.columns = [str(i) for i in raw_data.columns]

        return raw_data


class AutismChildDataset(Dataset):

    def __init__(self):
        super().__init__()
        self.name = 'autism_child'
        self.data_dir = os.path.join(DATA_DIR, self.name)

    def load_raw_data(self):
        raw_data = self.download_data()
        count_show_unique(raw_data)
        return raw_data

    def load(self):

        # download data
        raw_data = self.download_data()

        # specify meta data
        self.raw_data = raw_data
        self.sensitive_features = ['age', 'gender', 'ethnicity']
        self.drop_features = ['age_desc']
        self.ordinal_features = []
        self.multiclass_features = ['ethnicity', 'contry_of_res', 'relation']
        self.numerical_features = ['age', 'result']
        self.binary_features = [
            col for col in raw_data.columns 
            if col not in self.numerical_features + self.multiclass_features + self.ordinal_features + self.drop_features
        ]
        self.target_features = ['Class/ASD']

        # basic processing
        self.raw_data = self.basic_processing(raw_data)

    def handle_missing_data(self, data: pd.DataFrame):
        return data.dropna()

    def download_data(self):
        downloader = UCIMLDownloader(
            url = 'https://archive.ics.uci.edu/static/public/419/autistic+spectrum+disorder+screening+data+for+children.zip'
        )

        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        from scipy.io import arff

        raw_data, meta = arff.loadarff(os.path.join(self.data_dir, DATA_DOWNLOAD_DIR, 'Autism-Child-Data.arff'))
        
        raw_data = pd.DataFrame(raw_data)
        raw_data = raw_data.map(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x)
        raw_data.replace('?', np.nan, inplace = True)

        return raw_data
    

class AutismAdultDataset(Dataset):

    def __init__(self):
        super().__init__()
        self.name = 'autism_adult'
        self.data_dir = os.path.join(DATA_DIR, self.name)

    def load_raw_data(self):
        raw_data = self.download_data()
        count_show_unique(raw_data)
        return raw_data

    def load(self):

        # download data
        raw_data = self.download_data()

        # specify meta data
        self.raw_data = raw_data
        self.sensitive_features = ['age', 'gender', 'ethnicity']
        self.drop_features = ['age_desc']
        self.ordinal_features = []
        self.multiclass_features = ['ethnicity', 'contry_of_res', 'relation']
        self.numerical_features = ['age', 'result']
        self.binary_features = [
            col for col in raw_data.columns 
            if col not in self.numerical_features + self.multiclass_features + self.ordinal_features + self.drop_features
        ]
        self.target_features = ['Class/ASD']

        # basic processing
        self.raw_data = self.basic_processing(raw_data)

    def handle_missing_data(self, data: pd.DataFrame):
        return data.dropna()

    def download_data(self):
        downloader = UCIMLDownloader(
            url = 'https://archive.ics.uci.edu/static/public/426/autism+screening+adult.zip'
        )

        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        from scipy.io import arff

        raw_data, meta = arff.loadarff(os.path.join(self.data_dir, DATA_DOWNLOAD_DIR, 'Autism-Adult-Data.arff'))
        
        raw_data = pd.DataFrame(raw_data)
        raw_data = raw_data.map(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x)
        raw_data.replace('?', np.nan, inplace = True)

        return raw_data
    

class AutismAdolescentDataset(Dataset):

    def __init__(self):
        super().__init__()
        self.name = 'autism_adolescent'
        self.data_dir = os.path.join(DATA_DIR, self.name)

    def load_raw_data(self):
        raw_data = self.download_data()
        count_show_unique(raw_data)
        return raw_data

    def load(self):

        # download data
        raw_data = self.download_data()

        # specify meta data
        self.raw_data = raw_data
        self.sensitive_features = ['age', 'gender', 'ethnicity']
        self.drop_features = ['age_desc']
        self.ordinal_features = []
        self.multiclass_features = ['ethnicity', 'contry_of_res', 'relation']
        self.numerical_features = ['age', 'result']
        self.binary_features = [
            col for col in raw_data.columns 
            if col not in self.numerical_features + self.multiclass_features + self.ordinal_features + self.drop_features
        ]
        self.target_features = ['Class/ASD']

        # basic processing
        self.raw_data = self.basic_processing(raw_data)

    def handle_missing_data(self, data: pd.DataFrame):
        return data.dropna()

    def download_data(self):
        downloader = UCIMLDownloader(
            url = 'https://archive.ics.uci.edu/static/public/420/autistic+spectrum+disorder+screening+data+for+adolescent.zip'
        )

        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        from scipy.io import arff

        raw_data, meta = arff.loadarff(os.path.join(self.data_dir, DATA_DOWNLOAD_DIR, 'Autism-Adolescent-Data.arff'))
        
        raw_data = pd.DataFrame(raw_data)
        raw_data = raw_data.map(lambda x: x.decode('utf-8') if isinstance(x, bytes) else x)
        raw_data.replace('?', np.nan, inplace = True)

        return raw_data
    
    
class GENE3494Dataset(Dataset):

    def __init__(self, data_dir, k = 30):
        super().__init__()
        self.data_dir = data_dir
        self.name = 'GENE3494'
        self.k = k  # k - number of gene features to use from each platform
        self.data_dir = os.path.join(DATA_DIR, self.name)

    def load_raw_data(self):
        raw_data = self.download_data(self.data_dir)
        return raw_data

    def load(self):

        # download data
        raw_data = self.download_data(self.data_dir)

        # specify meta data
        self.raw_data = raw_data
        self.sensitive_features = ['age_at_diagnosis']
        self.drop_features = []
        self.ordinal_features = []
        self.binary_features = [
            'p53_seq_mut_status', 'p53.DLDA_classifier_result', 'PgR_status', 'DSS_EVENT'
        ]
        self.multiclass_features = [
            'DLDA_error', 'Elston_histologic_grade', 'ER_status', 'Lymph_node_status'
        ]
        
        self.target_features = [
            'DSS_EVENT'
        ]
        self.numerical_features = [
            col for col in raw_data.columns if col not in self.binary_features + self.multiclass_features + self.ordinal_features
        ]

        # basic processing
        self.raw_data = self.basic_processing(raw_data)

    def handle_missing_data(self, data: pd.DataFrame):
        return data.dropna()

    def download_data(self, data_dir: str):
        
        # load local files from data_dir
        df_gene96 = pd.read_csv(os.path.join(data_dir, 'genotype_data_gpl96.csv'))
        df_gene97 = pd.read_csv(os.path.join(data_dir, 'genotype_data_gpl97.csv'))
        ttest96 = pd.read_csv(os.path.join(data_dir, 'ttest_results_gpl96.csv'))
        ttest97 = pd.read_csv(os.path.join(data_dir, 'ttest_results_gpl97.csv'))
        df_pheno = pd.read_csv(os.path.join(data_dir, 'phenotype_data.csv'))
        
        ttest96.sort_values(by = 'p_value', inplace = True, ascending = True)
        feature_set96 = ttest96.loc[:self.k, 'feature'].tolist()
        
        ttest97.sort_values(by = 'p_value', inplace = True, ascending = True)
        feature_set97 = ttest97.loc[:self.k, 'feature'].tolist()
        
        df_gene96 = df_gene96[['ID'] + feature_set96]
        df_gene97 = df_gene97[['ID'] + feature_set97]
        df = pd.merge(df_gene96, df_gene97, on = 'ID')
        df = pd.merge(df, df_pheno, on = 'ID')
        df.drop(columns = ['ID'], inplace = True)
        
        phenotype_features = df_pheno.columns.tolist()
        phenotype_features.remove('ID')
        phenotype_features.remove('DSS_EVENT')
        self.feature_groups = {
            'phenotype': phenotype_features,
            'gene_96': feature_set96,
            'gene_97': feature_set97
        }
        
        raw_data = df

        return raw_data
    
    def get_ml_ready_data(self, task_name: str = None):
        
        from MedDataKit.data_pipe_routines.data_type_routines import basic_data_type_formulation, update_data_type_info
        from MedDataKit.data_pipe_routines.data_encoding_routines import basic_categorical_encoding
        from MedDataKit.data_pipe_routines.data_encoding_routines import basic_numerical_encoding
        from MedDataKit.data_pipe_routines.missing_data_routines import basic_missing_mitigation
        
        data = self.raw_data.copy()
        data_preprocessing_log = {
            'raw_data_shape': data.shape
        }
                
        if task_name is None:
            task_name = 'DSS_EVENT'
        
        if task_name == 'DSS_EVENT':
            
            # specify target and task type
            target = 'DSS_EVENT'
            task_type = 'classification'
            num_classes = data[target].nunique()
            
            # specify drop features
            drop_features = self.target_features.copy()
            drop_features.remove('DSS_EVENT')
        else:
            raise ValueError(f"Task {task_name} is not supported")
        
        ########################################################################################
        # drop features
        ########################################################################################
        data.drop(columns = drop_features, inplace = True)

        ########################################################################################
        # data type formulation
        ########################################################################################
        data, data_type_info = basic_data_type_formulation(
            data = data,
            numerical_cols = self.numerical_features,
            ordinal_cols = self.ordinal_features,
            binary_cols = self.binary_features,
            multiclass_cols = self.multiclass_features,
            target_col = target
        )
                
        ########################################################################################
        # missing data mitigation
        ########################################################################################
        data, missing_data_info = basic_missing_mitigation(data, threshold1 = 0.15)
        
        data_type_info = update_data_type_info(data, data_type_info, target)
        numerical_features = data_type_info['numerical_feature']
        categorical_features = data_type_info['categorical_feature']
        
        ########################################################################################
        # data encoding
        ########################################################################################
        data, numerical_encoding_info = basic_numerical_encoding(data, numerical_features)
        # print(data_encoding_info)
        # data, data_encoding_info = basic_categorical_encoding(data, categorical_features)
        # print(data_encoding_info)
        
        data_type_info = update_data_type_info(data, data_type_info, target)
        numerical_features = data_type_info['numerical_feature']
        categorical_features = data_type_info['categorical_feature']
        
        ########################################################################################
        # data configuration
        ########################################################################################
        data = data[numerical_features + categorical_features + [target]]
        data_config = {
            'numerical_feature': numerical_features,
            'categorical_feature': categorical_features,
            'target': target,
            'task_type': task_type,
            'num_classes': num_classes,
        }
        
        data_preprocessing_log.update({
            'data_type_info': data_type_info,
            'missing_data_info': missing_data_info,
            'numerical_encoding_info': numerical_encoding_info,
            'processed_data_shape': data.shape
        })
        
        ##########################################################################################
        # Logging
        ##########################################################################################
        print('Data preprocessing log:')
        print("Raw data shape: ", data_preprocessing_log['raw_data_shape'])  # todo: add more details
        print("Processed data shape: ", data_preprocessing_log['processed_data_shape'])
        
        return data, data_config, data_preprocessing_log
    



import pandas as pd
import os
import pyreadr
import numpy as np
from typing import Tuple, Optional, List, Dict, Any, Union
from config import DATA_DIR, DATA_DOWNLOAD_DIR
from scipy.io import arff

from ..downloader import OpenMLDownloader
from ..downloader import RDataDownloader
from ..downloader import KaggleDownloader
from ..downloader import UCIMLDownloader
from .base_dataset import Dataset
from .base_raw_dataset import RawDataset
from .base_ml_task_dataset import MLTaskDataset, MLTaskPreparationConfig
from ..data_pipe_routines.missing_data_routines import BasicMissingDataHandler
from ..data_pipe_routines.data_type_routines import BasicFeatureTypeHandler
from ..utils import handle_targets

###################################################################################################################################
# Arrhythmia Dataset
###################################################################################################################################
class ArrhythmiaDataset(Dataset):

    def __init__(self):
        
        name = 'arrhythmia'
        subject_area = 'Medical'
        year = 1997
        url = 'https://archive.ics.uci.edu/dataset/5/arrhythmia'
        download_link = 'https://archive.ics.uci.edu/static/public/5/arrhythmia.zip'
        description = "Distinguish between the presence and absence of cardiac arrhythmia and classify" \
                      "it in one of the 16 groups."
        notes = 'Extracted from Signals'
        source = 'uci'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
        
    def _load_raw_data(self):
        
        # download data
        downloader = UCIMLDownloader(url = self.download_link)
        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        # load raw data
        raw_data = pd.read_csv(
            os.path.join(self.data_dir, 'arrhythmia.data'), header = None, na_values='?'
        )
        
        columns = [
            'Age', 'Sex', 'Height', 'Weight', 'QRS duration', 'P-R interval', 'Q-T interval', 
            'T interval', 'P interval', 'QRS_angle', 'T_angle', 'P_angle', 'QRST_angle', 'J_angle', 'Heart rate'
        ]
        
        for key in ['DI', 'DII', 'DIII', 'AVR', 'AVL', 'AVF', 'V1', 'V2', 'V3', 'V4', 'V5', 'V6']:
            columns.append(f'{key}_Q')
            columns.append(f'{key}_R')
            columns.append(f'{key}_S')
            columns.append(f'{key}_R2')
            columns.append(f'{key}_S2')
            columns.append(f'{key}_num_deflections')
            columns.append(f'{key}_Ragged_R')
            columns.append(f'{key}_Diphasic_R')
            columns.append(f'{key}_Ragged_P')
            columns.append(f'{key}_Diphasic_P')
            columns.append(f'{key}_Ragged_T')
            columns.append(f'{key}_Diphasic_T')
        
        for key in ['DI', 'DII', 'DIII', 'AVR', 'AVL', 'AVF', 'V1', 'V2', 'V3', 'V4', 'V5', 'V6']:
            columns.append(f'{key}_Amp_JJ')
            columns.append(f'{key}_Amp_Q')
            columns.append(f'{key}_Amp_R')
            columns.append(f'{key}_Amp_S')
            columns.append(f'{key}_Amp_R2')
            columns.append(f'{key}_Amp_S2')
            columns.append(f'{key}_Amp_P')
            columns.append(f'{key}_Amp_T')
            columns.append(f'{key}_Amp_QRSA')
            columns.append(f'{key}_Amp_QRSTA')
        
        columns.append('Class')
        raw_data.columns = columns
        
        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
    
        # Specify meta data
        binary_features = ['Sex']
        for key in ['DI', 'DII', 'DIII', 'AVR', 'AVL', 'AVF', 'V1', 'V2', 'V3', 'V4', 'V5', 'V6']:
            binary_features.append(f'{key}_Ragged_R')
            binary_features.append(f'{key}_Diphasic_R')
            binary_features.append(f'{key}_Ragged_P')
            binary_features.append(f'{key}_Diphasic_P')
            binary_features.append(f'{key}_Ragged_T')
            binary_features.append(f'{key}_Diphasic_T')
        multiclass_features = ['Class']
        ordinal_features = []
        ordinal_feature_order_dict = {}
        numerical_features = [
            col for col in raw_data.columns 
            if col not in binary_features + multiclass_features + ordinal_features
        ]
        
        target_features = ['Class']
        sensitive_features = ['Age', 'Sex']
        drop_features = []
        task_names = ['Class']
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'fed_cols': fed_cols
        }

    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name 
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        if task_name == 'Class':
            
            target_info = {
                'target': 'Class',
                'task_type': 'classification'
            }
            
        else:
            raise ValueError(f"Invalid task name: {task_name}")
        
        assert (target_info['target'] in data.columns), "Target feature not found in data"
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        ordinal_as_numerical = ml_task_prep_config.ordinal_as_numerical
        feature_type_handler = BasicFeatureTypeHandler(ordinal_as_numerical)
        
        data, numerical_features, categorical_features = feature_type_handler.handle_feature_type(
            data, data_config
        )
        
        return data, {
            'numerical_features': numerical_features,
            'categorical_features': categorical_features,
        }
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        missing_data_handler = BasicMissingDataHandler(
            threshold1_num = 0.5, threshold2_num = 0.2, 
            threshold1_cat = 0.5, threshold2_cat = 0.2
        )
        data, missing_data_info = missing_data_handler.handle_missing_data(data, categorical_features)
        assert data.isna().sum().sum() == 0, "Missing data is not handled"
        return data, missing_data_info
    

###################################################################################################################################
# Colposcopy Dataset
###################################################################################################################################
class ColposcopyDataset(Dataset):

    def __init__(self):
        
        name = 'colposcopy'
        subject_area = 'Medical'
        year = 2017
        url = 'https://archive.ics.uci.edu/dataset/384/quality+assessment+of+digital+colposcopies'
        download_link = 'https://archive.ics.uci.edu/static/public/384/quality+assessment+of+digital+colposcopies.zip'
        description = "This dataset explores the subjective quality assessment of digital colposcopies."
        notes = 'Extracted from Signals'
        source = 'uci'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
    
    def _load_raw_data(self) -> Tuple[pd.DataFrame, dict]:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        downloader = UCIMLDownloader(url = self.download_link)
        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        file_dir = os.path.join(self.data_dir, 'Quality Assessment - Digital Colposcopy')
        green = pd.read_csv(os.path.join(file_dir, 'green.csv'))
        green['class'] = 'green'
        hinselmann = pd.read_csv(os.path.join(file_dir, 'hinselmann.csv'))
        hinselmann['class'] = 'hinselmann'
        schiller = pd.read_csv(os.path.join(file_dir, 'schiller.csv'))
        schiller['class'] = 'schiller'
        
        raw_data = pd.concat([green, hinselmann, schiller], axis=0)
        
        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        binary_features = [
            'experts::0', 'experts::1', 'experts::2', 'experts::3', 'experts::4', 'experts::5', 'consensus'
        ]
        ordinal_features = []
        ordinal_feature_order_dict = {}
        multiclass_features = ['class']
        numerical_features = [
            col for col in raw_data.columns 
            if col not in binary_features + ordinal_features + multiclass_features
        ]
        
        target_features = [
            'class', 'experts::0', 'experts::1', 'experts::2', 'experts::3', 'experts::4', 'experts::5', 'consensus'
        ]
        sensitive_features = []
        drop_features = []
        task_names = ['type_prediction', 'consensus_prediction']
        for i in range(6):
            task_names.append(f'expert_{i}_prediction')
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
    
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        
        if task_name == 'type_prediction':
            target_info = {
                'target': 'class',
                'task_type': 'classification'
            }
            
        elif task_name == 'consensus_prediction':
            target_info = {
                'target': 'consensus',
                'task_type': 'classification'
            }
        else:
            expert_num = int(task_name.split('_')[1])
            target_info = {
                'target': f'experts::{expert_num}',
                'task_type': 'classification'
            }
        
        data = handle_targets(data, raw_data_config, drop_unused_targets, target_info['target'])
        
        assert (target_info['target'] in data.columns), "Target feature not found in data"
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        
        """
        ordinal_as_numerical = ml_task_prep_config.ordinal_as_numerical
        feature_type_handler = BasicFeatureTypeHandler(ordinal_as_numerical)
        
        data, numerical_features, categorical_features = feature_type_handler.handle_feature_type(
            data, data_config
        )
        
        return data, {
            'numerical_features': numerical_features,
            'categorical_features': categorical_features,
        }
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        return data.dropna(), {}
    
###################################################################################################################################
# SPECTF Dataset
###################################################################################################################################
class SPECTFDataset(Dataset):

    def __init__(self):
        
        name = 'spectf'
        subject_area = 'Medical'
        year = 2001
        url = 'https://archive.ics.uci.edu/dataset/96/spectf+heart'
        download_link = 'https://archive.ics.uci.edu/static/public/96/spectf+heart.zip'
        description = "Data on cardiac Single Proton Emission Computed Tomography (SPECT) images." \
                      "Each patient classified into two categories: normal and abnormal."
        notes = 'Extracted from Signals'
        data_type = 'numerical'
        source = 'uci'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            data_type = data_type,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
    
    def _load_raw_data(self) -> pd.DataFrame:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        # download data
        downloader = UCIMLDownloader(url = self.download_link)
        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        # load raw data
        train = pd.read_csv(os.path.join(self.data_dir, 'SPECTF.train'), header = None).reset_index(drop = True)
        test = pd.read_csv(os.path.join(self.data_dir, 'SPECTF.test'), header = None).reset_index(drop = True)
        raw_data = pd.concat([train, test], axis = 0)
        
        columns = ['Diagnosis']
        for i in range(1, 23):
            columns.append(f'F{i}R')
            columns.append(f'F{i}S')
        raw_data.columns = columns
        
        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        binary_features = ['Diagnosis']
        ordinal_features = []
        ordinal_feature_order_dict = {}
        multiclass_features = []
        numerical_features = [
            col for col in raw_data.columns 
            if col not in binary_features + ordinal_features + multiclass_features
        ]
        
        target_features = ['Diagnosis']
        sensitive_features = []
        drop_features = []
        task_names = ['Diagnosis']
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
    
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """

        if task_name != 'Diagnosis':
            raise ValueError(f"task_name is not 'Diagnosis' for {self.name} dataset, which is not supported")
        
        target_info = {
            'target': 'Diagnosis',
            'task_type': 'classification'
        }
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        return data, {
            'numerical_features': data_config['numerical_features'],
            'categorical_features': [],
        }
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        return data.dropna(), {}
    
###################################################################################################################################
# Breast Cancer Wisconsin Dataset
###################################################################################################################################
class BreastCancerWisconsinDataset(Dataset):

    def __init__(self):
        
        name = 'breast_wisc'
        subject_area = 'Medical'
        year = 1995
        url = 'https://archive.ics.uci.edu/dataset/17/breast+cancer+wisconsin+diagnostic'
        download_link = 'https://archive.ics.uci.edu/static/public/17/breast+cancer+wisconsin+diagnostic.zip'
        description = "Diagnostic Wisconsin Breast Cancer Database."
        notes = 'Extracted from Signals'
        data_type = 'numerical'
        source = 'uci'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            data_type = data_type,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
    
    def _load_raw_data(self) -> pd.DataFrame:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        # download data
        downloader = UCIMLDownloader(url = self.download_link)
        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        # load raw data
        raw_data = pd.read_csv(os.path.join(self.data_dir, 'wdbc.data'), header = None, index_col = 0)
        raw_data.columns = ['diagnosis'] + [f'F{i-1}' for i in raw_data.columns[1:]]
        
        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        binary_features = ['diagnosis']
        ordinal_features = []
        ordinal_feature_order_dict = {}
        multiclass_features = []
        numerical_features = [
            col for col in raw_data.columns 
            if col not in binary_features + ordinal_features + multiclass_features
        ]
        
        target_features = ['diagnosis']
        sensitive_features = []
        drop_features = []
        task_names = ['diagnosis']
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
    
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """

        if task_name != 'diagnosis':
            raise ValueError(f"task_name is not 'diagnosis' for {self.name} dataset, which is not supported")
        
        target_info = {
            'target': 'diagnosis',
            'task_type': 'classification'
        }
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        return data, {
            'numerical_features': data_config['numerical_features'],
            'categorical_features': [],
        }
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        return data.dropna(), {}
    

###################################################################################################################################
# Dermatology Dataset
###################################################################################################################################
class DermatologyDataset(Dataset):

    def __init__(self):
        
        name = 'dermatology'
        subject_area = 'Medical'
        year = 1997
        url = 'https://archive.ics.uci.edu/dataset/33/dermatology'
        download_link = 'https://archive.ics.uci.edu/static/public/33/dermatology.zip'
        description = "Aim for this dataset is to determine the type of Eryhemato-Squamous Disease."
        notes = ''
        data_type = 'mixed'
        source = 'uci'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            data_type = data_type,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
    
    def _load_raw_data(self) -> pd.DataFrame:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        # download data
        downloader = UCIMLDownloader(url = self.download_link)
        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        # load raw data
        raw_data = pd.read_csv(os.path.join(self.data_dir, 'dermatology.data'), header=None, na_values='?')
        column_names = [
            'erythema', 'scaling', 'definite_borders', 'itching', 'koebner_phenomenon', 'polygonal_papules', 
            'follicular_papules', 'oral_mucosal_involvement', 'knee_and_elbow_involvement', 'scalp_involvement',
            'family_history', 'melanin_incontinence', 'eosinophils_in_the_infiltrate', 'PNL_infiltrate',
            'fibrosis_of_the_papillary_dermis', 'exocytosis', 'acanthosis', 'hyperkeratosis', 'parakeratosis',
            'clubbing_of_the_rete_ridges', 'elongation_of_the_rete_ridges', 'thinning_of_the_suprapapillary_epidermis',
            'spongiform_pustule', 'munro_microabcess', 'focal_hypergranulosis', 'disappearance_of_the_granular_layer',
            'vacuolisation_and_damage_of_basal_layer', 'spongiosis', 'saw_tooth_appearance_of_retes', 'follicular_horn_plug',
            'perifollicular_parakeratosis', 'inflammatory_monoluclear_inflitrate', 'band_like_infiltrate', 'Age', 'Class'
        ]
        raw_data.columns = column_names
        
        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        binary_features = ['family_history']
        numerical_features = ['Age']
        ordinal_feature_order_dict = {}
        multiclass_features = ['Class']
        ordinal_features = [
            col for col in raw_data.columns 
            if col not in binary_features + numerical_features + multiclass_features
        ]
        for col in ordinal_features:
            ordinal_feature_order_dict[col] = [0, 1, 2, 3]
        
        target_features = ['Class']
        sensitive_features = []
        drop_features = []
        task_names = ['Class']
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
    
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """

        if task_name != 'Class':
            raise ValueError(f"task_name is not 'Class' for {self.name} dataset, which is not supported")
        
        target_info = {
            'target': 'Class',
            'task_type': 'classification'
        }
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        ordinal_as_numerical = ml_task_prep_config.ordinal_as_numerical
        feature_type_handler = BasicFeatureTypeHandler(ordinal_as_numerical)
        
        data, numerical_features, categorical_features = feature_type_handler.handle_feature_type(
            data, data_config
        )
        
        return data, {
            'numerical_features': numerical_features,
            'categorical_features': categorical_features,
        }
        
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        return data.dropna(), {}
    
###################################################################################################################################
# Bone Transplant Dataset
###################################################################################################################################
class BoneTransplantDataset(Dataset):

    def __init__(self):
        
        name = 'bonetransplant'
        subject_area = 'Medical'
        year = 2020
        url = 'https://archive.ics.uci.edu/dataset/565/bone+marrow+transplant+children?spm=wolai.workspace.0.0.5fd9761cERlBBA'
        download_link = 'https://archive.ics.uci.edu/static/public/565/bone+marrow+transplant+children.zip'
        description = "The data set describes pediatric patients with several hematologic diseases, " \
            "who were subject to the unmanipulated allogeneic unrelated donor hematopoietic stem cell transplantation."
        notes = ''
        data_type = 'mixed'
        source = 'uci'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            data_type = data_type,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
    
    def _load_raw_data(self) -> pd.DataFrame:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        # download data
        downloader = UCIMLDownloader(url = self.download_link)
        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        # load raw data
        data, meta = arff.loadarff(os.path.join(self.data_dir, 'bone-marrow.arff'))
        df = pd.DataFrame(data)
        for col in df.columns:
            df[col] = df[col].replace(b'?', pd.NA)
            try:
                # Decode bytes and convert to numeric, keeping NaN values
                df[col] = pd.to_numeric(df[col].str.decode('utf-8'), errors='coerce')
            except (ValueError, AttributeError):
                continue
        df.replace('?', np.nan, inplace=True)
        df = df[df['ANCrecovery'] < 1000000]
        df = df.drop(columns = ['Disease'])
        raw_data = df
        
        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        ordinal_features = []
        ordinal_feature_order_dict = {}
        numerical_features = [
            'Donorage', 'Recipientage', 'CD34kgx10d6', 'CD3dCD34', 'CD3dkgx10d8', 'ANCrecovery', 
            'PLTrecovery', 'Rbodymass', 'time_to_aGvHD_III_IV', 'survival_time'
        ]
        multiclass_features = ['DonorABO', 'RecipientABO', 'CMVstatus', 'HLAmatch', 'Antigen', 'Alel', 'HLAgrI']
        binary_features = [
            col for col in raw_data.columns 
            if col not in numerical_features + multiclass_features + ordinal_features
        ]
        
        target_features = ['survival_status']
        sensitive_features = ['Recipientgender']
        drop_features = []
        task_names = ['predict_survival', 'predict_survival_status']
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
    
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """

        if task_name == 'predict_survival':
            data['survival_encoded'] = data.apply(
                lambda row: f"{row['survival_status']}|{row['survival_time']}", axis = 1
            )
            data = data.drop(columns = ['survival_status', 'survival_time'])
            target_info = {
                'target': 'survival_encoded',
                'task_type': 'survival'
            }
        elif task_name == 'predict_survival_status':
            target_info = {
                'target': 'survival_status',
                'task_type': 'classification'
            }
        
        assert target_info['target'] in data.columns, f"target {target_info['target']} is not in data columns"
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        ordinal_as_numerical = ml_task_prep_config.ordinal_as_numerical
        feature_type_handler = BasicFeatureTypeHandler(ordinal_as_numerical)
        
        data, numerical_features, categorical_features = feature_type_handler.handle_feature_type(
            data, data_config
        )
        
        return data, {
            'numerical_features': numerical_features,
            'categorical_features': categorical_features,
        }
        
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        missing_data_handler = BasicMissingDataHandler(
            threshold1_num=0.5,
            threshold1_cat=0.5,
            threshold2_num=0.05,
            threshold2_cat=0.05
        )
        data, missing_data_info = missing_data_handler.handle_missing_data(data, categorical_features)
        return data, missing_data_info
    
###################################################################################################################################
# Parkinsons Dataset
###################################################################################################################################
class ParkinsonsDataset(Dataset):

    def __init__(self):
        
        name = 'parkinsons'
        subject_area = 'Medical'
        year = 2020
        url = 'https://archive.ics.uci.edu/dataset/174/parkinsons'
        download_link = 'https://archive.ics.uci.edu/static/public/174/parkinsons.zip'
        description = "Oxford Parkinson's Disease Detection Dataset"
        notes = 'Extracted from Signal'
        data_type = 'numerical'
        source = 'uci'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            data_type = data_type,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
    
    def _load_raw_data(self) -> pd.DataFrame:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        # download data
        downloader = UCIMLDownloader(url = self.download_link)
        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        # load raw data        
        raw_data = pd.read_csv(
            os.path.join(self.data_dir, 'parkinsons.data'), index_col = 0
        )
        raw_data.columns = [str(i) for i in raw_data.columns]

        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        ordinal_features = []
        ordinal_feature_order_dict = {}
        multiclass_features = []
        binary_features = ['status']
        numerical_features = [
            col for col in raw_data.columns 
            if col not in binary_features + multiclass_features + ordinal_features
        ]
        
        target_features = ['status']
        sensitive_features = []
        drop_features = []
        task_names = ['predict_status']
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
    
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """

        if task_name == 'predict_status':
            target_info = {
                'target': 'status',
                'task_type': 'classification'
            }
        
        assert target_info['target'] in data.columns, f"target {target_info['target']} is not in data columns"
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        return data, {
            'numerical_features': data_config['numerical_features'],
            'categorical_features': [],
        }
        
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        return data.dropna(), {}
    

###################################################################################################
# Cervical Risk Dataset
###################################################################################################
class CervicalRiskDataset(Dataset):

    def __init__(self):
        
        name = 'cervical_risk'
        subject_area = 'Medical'
        year = 2023
        url = 'https://archive.ics.uci.edu/dataset/383/cervical+cancer+risk+factors?'
        download_link = 'https://archive.ics.uci.edu/static/public/383/cervical+cancer+risk+factors.zip'
        description = "This dataset focuses on the prediction of indicators/diagnosis of cervical cancer. " \
                      "The features cover demographic information, habits, and historic medical records."
        notes = ''
        data_type = 'mixed'
        source = 'uci'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            data_type = data_type,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
    
    def _load_raw_data(self) -> pd.DataFrame:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        # download data
        downloader = UCIMLDownloader(url = self.download_link)
        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        # load raw data        
        raw_data = pd.read_csv(os.path.join(self.data_dir, 'risk_factors_cervical_cancer.csv'), na_values='?')
        
        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        ordinal_features = []
        ordinal_feature_order_dict = {}
        multiclass_features = []
        binary_features = ['status']
        numerical_features = [
            'Age', 'Number of sexual partners', 'First sexual intercourse', 'Num of pregnancies', 'Smokes (years)',
            'Smokes (packs/year)', 'Hormonal Contraceptives (years)', 'IUD (years)', 'STDs (number)',
            'STDs: Number of diagnosis', 'STDs: Time since first diagnosis', 'STDs: Time since last diagnosis',
        ]
        
        binary_features = [
            col for col in raw_data.columns 
            if col not in numerical_features + multiclass_features + ordinal_features
        ]
        
        target_features = ['Biopsy']
        sensitive_features = ['Age']
        drop_features = []
        task_names = ['predict_biopsy']
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
    
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """

        if task_name == 'predict_biopsy':
            target_info = {
                'target': 'Biopsy',
                'task_type': 'classification'
            }
        else:
            raise ValueError(f"task name {task_name} is not supported")
        
        assert target_info['target'] in data.columns, f"target {target_info['target']} is not in data columns"
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        ordinal_as_numerical = ml_task_prep_config.ordinal_as_numerical
        feature_type_handler = BasicFeatureTypeHandler(ordinal_as_numerical)
        
        data, numerical_features, categorical_features = feature_type_handler.handle_feature_type(
            data, data_config
        )
        
        return data, {
            'numerical_features': numerical_features,
            'categorical_features': categorical_features,
        }
        
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        missing_data_handler = BasicMissingDataHandler(
            threshold1_num = 0.5,
            threshold2_num = 0.05,
            threshold1_cat = 0.5,
            threshold2_cat = 0.05,
            impute_num = 'mean',
            impute_cat = 'other'
        )
        data, missing_data_info = missing_data_handler.handle_missing_data(data, categorical_features)
        
        return data, missing_data_info

###################################################################################################################################
# Bacteremia Dataset
###################################################################################################################################
class BacteremiaDataset(Dataset):

    def __init__(self):
        
        name = 'bacteremia'
        subject_area = 'Medical'
        year = 2020
        url = 'https://hbiostat.org/data/'
        download_link = 'https://zenodo.org/api/records/7554815/files-archive'
        description = "Bacteremia dataset"
        notes = 'Bacteremia'
        data_type = 'numerical'
        self.pub_link = 'A Risk Prediction Model for Screening Bacteremic Patients: A Cross Sectional Study; ' \
                        'Regression with Highly Correlated Predictors: Variable Omission Is Not the Solution'
        source = 'vdb'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            data_type = data_type,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
    
    def _load_raw_data(self) -> pd.DataFrame:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        # download data
        if not os.path.exists(os.path.join(self.data_dir, 'Bacteremia_public_S2.csv')):
            downloader = URLDownloader(url = self.download_link, zipfile = True)
            download_status = downloader._custom_download(data_dir = self.data_dir)
            if not download_status:
                raise Exception(f'Failed to download data for {self.name}')
        
        # load raw data
        raw_data = pd.read_csv(os.path.join(self.data_dir, 'Bacteremia_public_S2.csv'), index_col=0).reset_index(drop=True)
        
        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        ordinal_features = []
        ordinal_feature_order_dict = {}
        binary_features = [
            "SEX", "BloodCulture"
        ]
        multiclass_features = []
        
        numerical_features = [
            col for col in raw_data.columns 
            if col not in binary_features + multiclass_features + ordinal_features
        ]
        
        target_features = ['BloodCulture']
        sensitive_features = ["SEX", "AGE"]
        drop_features = []
        task_names = ['predict_BloodCulture']
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
    
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """

        if task_name == 'predict_BloodCulture':
            target_info = {
                'target': 'BloodCulture',
                'task_type': 'classification'
            }
        else:
            raise ValueError(f"task name {task_name} is not supported")
        
        assert target_info['target'] in data.columns, f"target {target_info['target']} is not in data columns"
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        ordinal_as_numerical = ml_task_prep_config.ordinal_as_numerical
        feature_type_handler = BasicFeatureTypeHandler(ordinal_as_numerical)
        
        data, numerical_features, categorical_features = feature_type_handler.handle_feature_type(
            data, data_config
        )
        
        return data, {
            'numerical_features': numerical_features,
            'categorical_features': categorical_features,
        }
        
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        missing_data_handler = BasicMissingDataHandler(
            threshold1_num = 0.4,
            threshold2_num = 0.05,
            threshold1_cat = 0.4,
            threshold2_cat = 0.05,
            impute_num = 'mean',
            impute_cat = 'other'
        )
        data, missing_data_info = missing_data_handler.handle_missing_data(data, categorical_features)
        
        return data, missing_data_info
    

###################################################################################################################################
# Fetal CAD Dataset
###################################################################################################################################
class FetalCTGDataset(Dataset):

    def __init__(self):
        
        name = 'fetalctg'
        subject_area = 'Medical'
        year = 2019
        url = 'https://www.kaggle.com/datasets/akshat0007/fetalhr'
        download_link = None
        description = "This dataset explores the subjective quality assessment of digital colposcopies."
        notes = 'Extracted from Signals, CTG'
        data_type = 'numerical'
        source = 'kaggle'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
    
    def _load_raw_data(self) -> Tuple[pd.DataFrame, dict]:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        downloader = KaggleDownloader(
            dataset_name = 'akshat0007/fetalhr',
            file_names = ['CTG.csv'],
            download_all = True
        )
        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        raw_data = pd.read_csv(os.path.join(self.data_dir, 'CTG.csv')).reset_index(drop=True)
        raw_data = raw_data.drop(columns = ['FileName', 'Date', 'SegFile'])
        raw_data = raw_data.dropna()
        
        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        binary_features = [
            'A', 'B', 'C', 'D', 'E', 'AD', 'DE', 'LD', 'FS', 'SUSP'
        ]
        ordinal_features = []
        ordinal_feature_order_dict = {}
        multiclass_features = ['CLASS', 'NSP']
        numerical_features = [
            col for col in raw_data.columns 
            if col not in binary_features + ordinal_features + multiclass_features
        ]
        
        target_features = [
            'A', 'B', 'C', 'D', 'E', 'AD', 'DE', 'LD', 'FS', 'SUSP',
            'CLASS', 'NSP'
        ]
        sensitive_features = []
        drop_features = []
        task_names = [
            'predict_pattern_A', 
            'predict_pattern_B', 
            'predict_pattern_C', 
            'predict_pattern_D', 
            'predict_pattern_E', 
            'predict_pattern_AD', 
            'predict_pattern_DE', 
            'predict_pattern_LD', 
            'predict_pattern_FS', 
            'predict_pattern_SUSP',
            'predict_pattern_CLASS',
            'predict_pattern_NSP'
        ]
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
    
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        
        if drop_unused_targets is False:
            raise ValueError(f"drop_unused_targets is False, which is not supported for this dataset.")
               
        target_name = task_name.split('_')[-1]
        target_info = {
            'target': target_name,
            'task_type': 'classification'
        }
        
        data = handle_targets(data, raw_data_config, drop_unused_targets, target_info['target'])
        
        assert (target_info['target'] in data.columns), "Target feature not found in data."
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        
        """
        ordinal_as_numerical = ml_task_prep_config.ordinal_as_numerical
        feature_type_handler = BasicFeatureTypeHandler(ordinal_as_numerical)
        
        data, numerical_features, categorical_features = feature_type_handler.handle_feature_type(
            data, data_config
        )
        
        return data, {
            'numerical_features': numerical_features,
            'categorical_features': categorical_features,
        }
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        return data.dropna(), {}


import pandas as pd
import os
import pyreadr
import numpy as np
from typing import Tuple, Optional, List, Dict, Any, Union
from config import DATA_DIR, DATA_DOWNLOAD_DIR
from scipy.io import arff

from ..downloader import OpenMLDownloader
from ..downloader import RDataDownloader
from ..downloader import KaggleDownloader
from ..downloader import UCIMLDownloader
from .base_dataset import Dataset
from .base_raw_dataset import RawDataset
from .base_ml_task_dataset import MLTaskDataset, MLTaskPreparationConfig
from ..data_pipe_routines.missing_data_routines import BasicMissingDataHandler
from ..data_pipe_routines.data_type_routines import BasicFeatureTypeHandler
from ..utils import handle_targets


def count_show_unique(data):
    for col in data.columns:
        unique_values = data[col].unique().tolist()
        data_type = data[col].dtype
        num_unique = data[col].nunique()
        if num_unique < 20:
            print(f"{col} ({data_type}) => {num_unique} values: {unique_values}")
        else:
            print(f"{col} ({data_type}) => {num_unique} values")


###################################################################################################
# Codon Dataset
###################################################################################################
class CodonUsageDataset(Dataset):

    def __init__(self):
        
        name = 'codon'
        subject_area = 'Medical'
        year = 2020
        url = 'https://archive.ics.uci.edu/dataset/577/codon+usage'
        download_link = 'https://archive.ics.uci.edu/static/public/577/codon+usage.zip'
        description = "DNA codon usage frequencies of a large sample of diverse biological organisms from different taxa"
        notes = 'Bioinformatics'
        data_type = 'numerical'
        source = 'uci'
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            data_type = data_type,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
    
    def _load_raw_data(self) -> pd.DataFrame:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        # download data
        downloader = UCIMLDownloader(url = self.download_link)
        download_status = downloader._custom_download(data_dir = self.data_dir)
        if not download_status:
            raise Exception(f'Failed to download data for {self.name}')
        
        # load raw data        
        raw_data = pd.read_csv(os.path.join(self.data_dir, 'codon_usage.csv'), sep=',', low_memory=False)
        raw_data = raw_data.drop(['SpeciesID', 'Ncodons', 'SpeciesName', 'DNAtype'], axis=1)
        
        raw_data.replace("non-B hepatitis virus", np.nan, inplace=True)
        raw_data.replace("12;I", np.nan, inplace=True)
        raw_data.replace('-', np.nan, inplace=True)

        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        ordinal_features = []
        ordinal_feature_order_dict = {}
        multiclass_features = ['Kingdom']
        binary_features = []
        numerical_features = [
            col for col in raw_data.columns 
            if col not in binary_features + multiclass_features + ordinal_features
        ]
        
        target_features = ['Kingdom']
        sensitive_features = []
        drop_features = []
        task_names = ['predict_kingdom']
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
    
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """

        if task_name == 'predict_kingdom':
            target_info = {
                'target': 'Kingdom',
                'task_type': 'classification'
            }
        
        assert target_info['target'] in data.columns, f"target {target_info['target']} is not in data columns"
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        return data, {
            'numerical_features': data_config['numerical_features'],
            'categorical_features': [],
        }
        
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        return data.dropna(), {}

 
###################################################################################################################################
# GENE3494 Dataset
###################################################################################################################################
class GENE3494Dataset(Dataset):

    def __init__(self, k = 30):
        
        name = 'gene3494'
        subject_area = 'Medical'
        year = 2020
        url = 'GEO'
        download_link = ''
        description = "GEO Dataset 3494"
        notes = 'Bioinformatics'
        data_type = 'numerical'
        source = 'geo'
        self.k = k  # k - number of gene features to use from each platform
        
        super().__init__(
            name = name,
            description = description,
            collection_year = year,
            subject_area = subject_area,
            url = url,
            download_link = download_link,
            notes = notes,
            data_type = data_type,
            source = source
        )
        
        self.data_dir = os.path.join(DATA_DIR, self.name)
        self.raw_dataset: RawDataset = None
        self.ml_ready_dataset: MLReadyDataset = None
        
    def _load_raw_data(self) -> pd.DataFrame:
        """
        Load raw dataset and specify meta data information
        
        Returns:
            raw_data: pd.DataFrame, raw data
            meta_data: dict, meta data
        """
        # load local files from data_dir
        df_gene96 = pd.read_csv(os.path.join(self.data_dir, 'genotype_data_gpl96.csv'))
        df_gene97 = pd.read_csv(os.path.join(self.data_dir, 'genotype_data_gpl97.csv'))
        ttest96 = pd.read_csv(os.path.join(self.data_dir, 'ttest_results_gpl96.csv'))
        ttest97 = pd.read_csv(os.path.join(self.data_dir, 'ttest_results_gpl97.csv'))
        df_pheno = pd.read_csv(os.path.join(self.data_dir, 'phenotype_data.csv'))
        
        ttest96.sort_values(by = 'p_value', inplace = True, ascending = True)
        feature_set96 = ttest96.loc[:self.k, 'feature'].tolist()
        
        ttest97.sort_values(by = 'p_value', inplace = True, ascending = True)
        feature_set97 = ttest97.loc[:self.k, 'feature'].tolist()
        
        df_gene96 = df_gene96[['ID'] + feature_set96]
        df_gene97 = df_gene97[['ID'] + feature_set97]
        df = pd.merge(df_gene96, df_gene97, on = 'ID')
        df = pd.merge(df, df_pheno, on = 'ID')
        df.drop(columns = ['ID'], inplace = True)
        
        phenotype_features = df_pheno.columns.tolist()
        phenotype_features.remove('ID')
        phenotype_features.remove('DSS_EVENT')
        self.feature_groups = {
            'phenotype': phenotype_features,
            'gene_96': feature_set96,
            'gene_97': feature_set97
        }
        
        raw_data = df

        return raw_data
    
    def _set_raw_data_config(self, raw_data: pd.DataFrame) -> dict:
        """
        Set raw data configuration
        
        Returns:
            raw_data_config: dict, raw data configuration
        """
        ordinal_features = []
        ordinal_feature_order_dict = {}
        binary_features = [
            'p53_seq_mut_status', 'p53.DLDA_classifier_result', 'PgR_status', 'DSS_EVENT'
        ]
        multiclass_features = [
            'DLDA_error', 'Elston_histologic_grade', 'ER_status', 'Lymph_node_status'
        ]
        
        numerical_features = [
            col for col in raw_data.columns if col not in binary_features + multiclass_features + ordinal_features
        ]
        
        target_features = ['DSS_EVENT']
        sensitive_features = ['age_at_diagnosis']
        drop_features = []
        task_names = ['predict_dss_event']
        
        feature_groups = {}
        fed_cols = []
        
        return {
            'numerical_features': numerical_features,
            'binary_features': binary_features,
            'multiclass_features': multiclass_features,
            'ordinal_features': ordinal_features,
            'target_features': target_features,
            'sensitive_features': sensitive_features,
            'drop_features': drop_features,
            'feature_groups': feature_groups,
            'task_names': task_names,
            'ordinal_feature_order_dict': ordinal_feature_order_dict,
            'fed_cols': fed_cols
        }
        
    def _set_target_feature(
        self, data: pd.DataFrame, raw_data_config: dict, task_name: str, drop_unused_targets: bool
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            drop_unused_targets: bool, whether to drop unused target features
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """

        if task_name == 'predict_dss_event':
            target_info = {
                'target': 'DSS_EVENT',
                'task_type': 'classification'
            }
        
        assert target_info['target'] in data.columns, f"target {target_info['target']} is not in data columns"
        
        return data, target_info
    
    def _feature_engineering(
        self, data: pd.DataFrame, data_config: dict, ml_task_prep_config: MLTaskPreparationConfig = None
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Set target feature based on task name
        
        Args:
            data: pd.DataFrame, raw data
            raw_data_config: dict, raw data configuration {'target', 'task_type'}
            task_name: str, task name
            
        Returns:
            data: pd.DataFrame, processed data
            target_info: dict, target information
        """
        return data, {
            'numerical_features': data_config['numerical_features'],
            'categorical_features': [],
        }
        
    
    def _handle_missing_data(self, data: pd.DataFrame, categorical_features: list) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        
        Args:
            data: pd.DataFrame, raw data
            categorical_features: list, categorical features
            
        Returns:
            data: pd.DataFrame, processed data
            missing_data_info: dict, missing data processing information
        """
        return data.dropna(), {}


from sklearn.datasets import fetch_openml
import os
import pandas as pd
from abc import ABC, abstractmethod

import tempfile
import tarfile
import zipfile
import os
import pandas as pd
import requests
import io
import pyreadr
import shutil
import requests
import zipfile
import io
import os
import kaggle as kg
from dotenv import load_dotenv

from config import DATA_DIR, DATA_DOWNLOAD_DIR


class DownLoader(ABC):

    def __init__(self):
        pass

    def download(self, data_save_dir: str):

        # check if the data file exists
        if os.path.exists(data_save_dir):
            # if the file exists, load and return the data frame
            data = pd.read_csv(data_save_dir)
            return data
        else:
            # if the file does not exist, call the custom download function
            data = self._custom_download()
            
            # ensure the target directory exists
            os.makedirs(os.path.dirname(data_save_dir), exist_ok=True)
            
            # save the data to the specified path
            data.to_csv(os.path.join(data_save_dir, 'raw_data.csv'), index=False)
            
            return data

    @abstractmethod
    def _custom_download(self):
        """
        Customized download function to download the data from different sources
        """
        pass


class UCIMLDownloader(DownLoader):

    def __init__(self, url: str):
        self.url = url
        super().__init__()

    def _custom_download(self, data_dir: str):

        try:
            download_dir = os.path.join(data_dir)
            if not os.path.exists(download_dir):
                os.makedirs(download_dir)

            url = self.url
            zipfile_name = url.split('/')[-1]
            zipfile_path = os.path.join(download_dir, zipfile_name)
            # check if the zip file exists, if not, download and save zip file
            if not os.path.exists(zipfile_path):
                # download data and unzip to download_dir
                response = requests.get(url)
                
                # save response content to zipfile_path
                with open(zipfile_path, 'wb') as f:
                    f.write(response.content)

            # unzip the zip file
            with zipfile.ZipFile(zipfile_path, 'r') as zip_ref:
                zip_ref.extractall(download_dir)
                
            # return True if the download is successful
            return True
        except Exception as e:
            print(e)
            return False


class KaggleDownloader(DownLoader):
    
    """
    Downloader for Kaggle datasets
    
    Args:
        dataset_name: str, name of the dataset - e.g 'elhamnasarian/nasarian-cad-dataset
        file_names: list[str], names of the files to download
        download_all: bool, whether to download all files
    
    """

    def __init__(
        self, dataset_name: str, 
        file_names: list[str], 
        download_all: bool = False,
        competition: bool = False
    ):
        
        self.dataset_name = dataset_name
        self.file_names = file_names
        self.download_all = download_all
        self.competition = competition
        super().__init__()
        
    def _custom_download(self, data_dir: str):

        try:
            # load kaggle username and key from .env file
            dotenv_path = '.env'
            load_dotenv(dotenv_path)
            kaggle_username = os.getenv('KAGGLE_USERNAME')
            kaggle_key = os.getenv('KAGGLE_KEY')

            # init kaggle api by setting kaggle.json
            KaggleDownloader.init_on_kaggle(kaggle_username, kaggle_key)
            kg.api.authenticate()

            # download data based on the parameters    
            download_dir = os.path.join(data_dir)
            if not os.path.exists(download_dir):
                os.makedirs(download_dir)
            
            if not self.competition:
                dataset_name = self.dataset_name
                file_names = self.file_names
                download_all = self.download_all

                # download all files as whole
                if download_all:
                    zipfile_name = dataset_name.split('/')[-1] + '.zip'
                    zipfile_path = os.path.join(download_dir, zipfile_name)
                    if not os.path.exists(zipfile_name):
                        kg.api.dataset_download_files(dataset_name, path=download_dir, unzip = False)
                    
                    # unzip zip file
                    zip_file_path = os.path.join(download_dir, zipfile_name)
                    print(zip_file_path)
                    with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
                        zip_ref.extractall(download_dir)
                # download only specific files by file_names
                else:
                    for file_name in file_names:
                        if not os.path.exists(os.path.join(download_dir, file_name)):
                            kg.api.dataset_download_file(dataset_name, file_name, path=download_dir)
            else:
                competition_name = self.dataset_name
                file_names = self.file_names
                download_all = self.download_all

                if download_all:
                    zipfile_name = competition_name.split('/')[-1] + '.zip'
                    zipfile_path = os.path.join(download_dir, zipfile_name)
                    if not os.path.exists(zipfile_name):
                        kg.api.competition_download_files(competition_name, path=download_dir, quiet = False)
                    
                    # unzip zip file
                    zip_file_path = os.path.join(download_dir, zipfile_name)
                    print(zip_file_path)
                    with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
                        zip_ref.extractall(download_dir)
                else:
                    for file_name in file_names:
                        if not os.path.exists(os.path.join(download_dir, file_name)):
                            kg.api.competition_download_file(competition_name, file_name, path=download_dir)
                        
            return True
        except Exception as e:
            print(e)
            return False
        
    
    @staticmethod
    def init_on_kaggle(username, api_key):

        import os
        import json
        import subprocess

        # set kaggle config dir based on os
        if os.name == 'nt':
            # windows
            KAGGLE_CONFIG_DIR = os.path.join(os.path.expandvars('%USERPROFILE%'), '.kaggle')
        else:
            KAGGLE_CONFIG_DIR = os.path.join(os.path.expandvars('$HOME'), '.kaggle')
        
        os.makedirs(KAGGLE_CONFIG_DIR, exist_ok = True)
        api_dict = {"username":username, "key":api_key}
        file_path = os.path.join(KAGGLE_CONFIG_DIR, 'kaggle.json')
        print(file_path)
        with open(file_path, "w", encoding='utf-8') as f:
            json.dump(api_dict, f)
        
        # code for permission setting
        # if os.name == 'nt':
        #     # windows -> file path is different
        #     cmd = f"chmod 600 {file_path}"
        # else:
        #     cmd = f"chmod 600 {KAGGLE_CONFIG_DIR}/kaggle.json"
        # print(cmd)
        # output = subprocess.check_output(cmd.split(" "))
        # output = output.decode(encoding='UTF-8')
        # print(output)


class RDataDownloader(DownLoader):

    def __init__(self, package_url: str, dataset_path_in_package: str):
        self.package_url = package_url
        self.dataset_path_in_package = dataset_path_in_package
        super().__init__()

    def _custom_download(self, data_dir: str):
        
        download_dir = os.path.join(data_dir, DATA_DOWNLOAD_DIR)
        if not os.path.exists(download_dir):
            os.makedirs(download_dir)

        dataset_path = self.dataset_path_in_package.split('/')[-1]

        # if dataset_path exists, return True
        if os.path.exists(os.path.join(download_dir, dataset_path)):
            return True
        # if dataset_path does not exist, download the package and extract the dataset
        else:
            # create a temporary directory
            temp_dir = tempfile.mkdtemp()
            error_happened = False
            try:
                # download the package
                response = requests.get(self.package_url)
                tar_content = io.BytesIO(response.content)
                
                # extract the package
                with tarfile.open(fileobj=tar_content, mode='r:gz') as tar:
                    tar.extractall(path=temp_dir)
                
                # read the .rda file
                rda_path = os.path.join(temp_dir, self.dataset_path)

                # copy the .rda file to download_dir
                shutil.copy(rda_path, download_dir)
                
            except Exception as e:
                print(e)
                error_happened = True
            finally:
                # delete the temporary directory
                shutil.rmtree(temp_dir)
                return error_happened


class OpenMLDownloader(DownLoader):

    def __init__(self, data_id: int):
        self.data_id = data_id
        super().__init__()

    def _custom_download(self, data_dir: str):

        try:
            download_dir = os.path.join(data_dir, DATA_DOWNLOAD_DIR)
            if not os.path.exists(download_dir):
                os.makedirs(download_dir)

            X, y = fetch_openml(data_id=self.data_id, as_frame=True, return_X_y=True)
            data = X.join(y)

            # save the data to the specified path
            data.to_csv(os.path.join(download_dir, 'data.csv'), index=False)

            return True
        
        except Exception as e:
            print(e)
            return False
        
class URLDownloader(DownLoader):

    def __init__(self, url: str, zipfile: bool = False):
        self.url = url
        self.zipfile = zipfile
        super().__init__()

    def _custom_download(self, data_dir: str):

        try:
            download_dir = os.path.join(data_dir)
            if not os.path.exists(download_dir):
                os.makedirs(download_dir)

            url = self.url
            file_name = url.split('/')[-1]
            file_path = os.path.join(download_dir, file_name)
            # check if the file exists, if not, download and save file
            if not os.path.exists(file_path):
                # download data and unzip to download_dir
                response = requests.get(url)
                
                # save response content to file_path
                with open(file_path, 'wb') as f:
                    f.write(response.content)

            # unzip the zip file
            if self.zipfile:
                with zipfile.ZipFile(file_path, 'r') as zip_ref:
                    zip_ref.extractall(download_dir)
                
            # return True if the download is successful
            return True
        except Exception as e:
            print(e)
            return False
        
        
class LocalDownloader(DownLoader):

    def __init__(self, file_names: list[str]):
        self.file_names = file_names
        super().__init__()

    def _custom_download(self, data_dir: str):
        
        if not os.path.exists(data_dir):
            os.makedirs(data_dir)
        
        if not all(os.path.exists(os.path.join(data_dir, file_name)) for file_name in self.file_names):
            raise Exception(f'File {file_name} does not exist in {data_dir}')
        
        return True



from typing import Tuple
from abc import ABC, abstractmethod
import pandas as pd


def simple_impute(data: pd.Series, impute_strategy: str, is_categorical: bool = False) -> pd.Series:
    if is_categorical:
        if impute_strategy == 'mode':
            data = data.fillna(data.mode().iloc[0])
        elif impute_strategy == 'other':
            if data.dtype == 'object' or data.dtype == 'category':
                data = data.fillna('None')
            else:
                data = data.fillna(-999)
    else:
        if impute_strategy == 'mean':
            data = data.fillna(data.mean())
        elif impute_strategy == 'median':
            data = data.fillna(data.median())
        elif impute_strategy == 'zero':
            data = data.fillna(0)
        elif impute_strategy == 'other':
            data = data.fillna(-999)
    return data

class MissingDataHandler(ABC):
    
    def __init__():
        pass
    
    @abstractmethod
    def handle_missing_data(self, data: pd.DataFrame, categorical_features: list, **kwargs) -> Tuple[pd.DataFrame, dict]:
        pass
    

class BasicMissingDataHandler(MissingDataHandler):
    """
    Basic missing data mitigation routines
    1. if missing ratio > threshold1, drop the feature
    2. if missing ratio < threshold2, drop the row
    3. if missing ratio > threshold1 and < threshold2, impute with mean/median/mode
    """
    
    def __init__(
        self, 
        threshold1_num: float = 0.5, 
        threshold2_num: float = 0.1,
        threshold1_cat: float = 0.5, 
        threshold2_cat: float = 0.1,
        impute_num: str = 'mean',
        impute_cat: str = 'other',
    ):
        self.threshold1_num = threshold1_num
        self.threshold2_num = threshold2_num
        self.threshold1_cat = threshold1_cat
        self.threshold2_cat = threshold2_cat
        self.impute_num = impute_num
        self.impute_cat = impute_cat
        
        assert self.impute_num in ['mean', 'median', 'zero'], "Invalid impute_num"
        assert self.impute_cat in ['other', 'mode'], "Invalid impute_cat"
    
    def handle_missing_data(
        self, data: pd.DataFrame, categorical_features: list
    ) -> Tuple[pd.DataFrame, dict]:
        """
        Handle missing data
        """
        missing_data_info = {}
        for feature in data.columns:
            missing_ratio = data[feature].isnull().sum() / len(data)
            if feature in categorical_features:
                if missing_ratio > self.threshold1_cat:
                    data = data.drop(columns=[feature])
                    missing_data_info[feature] = {
                        'action': 'drop_col', 'missing_ratio': '{:.2f}'.format(missing_ratio), 
                    }
                elif missing_ratio > self.threshold2_cat:
                    data[feature] = simple_impute(data[feature], self.impute_cat, is_categorical=True)
                    missing_data_info[feature] = {
                        'action': 'impute', 'missing_ratio': '{:.2f}'.format(missing_ratio), 
                        'impute_value': self.impute_cat
                    }
                else:
                    missing_data_info[feature] = {'action': 'drop_row', 'missing_ratio': '{:.2f}'.format(missing_ratio)}
            else:
                if missing_ratio > self.threshold1_num:
                    data = data.drop(columns=[feature])
                    missing_data_info[feature] = {
                        'action': 'drop_col', 'missing_ratio': '{:.2f}'.format(missing_ratio), 
                    }
                elif missing_ratio > self.threshold2_num:
                    data[feature] = simple_impute(data[feature], self.impute_num, is_categorical=False)
                    missing_data_info[feature] = {
                        'action': 'impute', 'missing_ratio': '{:.2f}'.format(missing_ratio), 
                        'impute_value': self.impute_num
                    }
                else:
                    missing_data_info[feature] = {'action': 'drop_row', 'missing_ratio': '{:.2f}'.format(missing_ratio)}
        
        # drop rows with missing values for remaining features
        data = data.dropna()
                
        return data, missing_data_info
    
    
    
    



import pandas as pd
from typing import List, Tuple
import numpy as np
from abc import ABC, abstractmethod

from MedDataKit.utils import update_feature_type

def basic_data_type_formulation(
    data: pd.DataFrame,
    numerical_cols: List[str],
    ordinal_cols: List[str],
    binary_cols: List[str],
    multiclass_cols: List[str],
    target_col: str
) -> Tuple[pd.DataFrame, dict]:
    """
    Basic data formulation rountines: determine feature types, target variable and task type
    - numerical: ['numerical', 'ordinal']
    - categorical: ['binary', 'multiclass']
    - target: target feature
    - task: classification or regression
    """
    
    numerical_features = []
    categorical_features = []
    
    for col in data.columns:
        if col == target_col:
            continue
        
        if col in numerical_cols or col in ordinal_cols:
            numerical_features.append(col)
        else:
            categorical_features.append(col)
            
    # data type conversion
    data[numerical_features] = data[numerical_features].astype(float)
    data[categorical_features] = data[categorical_features].astype('Int64')
    print(categorical_features)
    
    assert target_col not in numerical_features and target_col not in categorical_features, "Target column cannot be numerical or categorical"
    
    data = data[numerical_features + categorical_features + [target_col]]
        

    return data, {
        'numerical_feature': numerical_features,
        'categorical_feature': categorical_features,
        'target': target_col
    }
    
    
def update_data_type_info(data: pd.DataFrame, data_type_info: dict, target_col: str) -> dict:
    """
    Update data type information
    """
    numerical_features = data_type_info['numerical_feature']
    categorical_features = data_type_info['categorical_feature']
    
    new_numerical_features = []
    new_categorical_features = []
    for col in data.columns:
        if col == target_col:
            continue
        
        if col in numerical_features:
            new_numerical_features.append(col)
        else:
            new_categorical_features.append(col)
    
    assert target_col not in new_numerical_features and target_col not in new_categorical_features, "Target column cannot be numerical or categorical"
    data_type_info['numerical_feature'] = new_numerical_features
    data_type_info['categorical_feature'] = new_categorical_features
    
    return data_type_info


class FeatureTypeHandler(ABC):
    
    def __init__(self):
        pass
    
    @abstractmethod
    def handle_feature_type(
        self, data: pd.DataFrame, data_config: dict, **kwargs
    ) -> Tuple[pd.DataFrame, list, list]:
        pass


class BasicFeatureTypeHandler(FeatureTypeHandler):
    
    def __init__(
        self,
        ordinal_as_numerical: bool,
    ):
        
        self.ordinal_as_numerical = ordinal_as_numerical
    
    def handle_feature_type(
        self, data: pd.DataFrame, data_config: dict, **kwargs
    ) -> Tuple[pd.DataFrame, list, list]:
        
        # determine numerical and categorical features
        if self.ordinal_as_numerical:
            numerical_features = data_config['numerical_features'] + data_config['ordinal_features']
            categorical_features = data_config['multiclass_features'] + data_config['binary_features']
        else:
            numerical_features = data_config['numerical_features']
            categorical_features = (
                data_config['multiclass_features'] + data_config['binary_features'] 
                + data_config['ordinal_features']
            )
        
        # remove target feature from numerical and categorical features
        target_feature = data_config['target']
        
        if target_feature in numerical_features:
            numerical_features.remove(target_feature)
        if target_feature in categorical_features:
            categorical_features.remove(target_feature)
        
        # remove other features from numerical and categorical features
        numerical_features = [col for col in numerical_features if col in data.columns]
        categorical_features = [col for col in categorical_features if col in data.columns]
        
        data[numerical_features] = data[numerical_features].astype(float)
        data[categorical_features] = data[categorical_features].astype('str')
        
        return data, numerical_features, categorical_features





import pandas as pd
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from typing import Tuple, List
from sklearn.preprocessing import OneHotEncoder

def basic_numerical_encoding(
    data: pd.DataFrame,
    numerical_features: List[str],
    standardize: bool = True,
    normalize: bool = True
) -> Tuple[pd.DataFrame, dict]:
    """
    Basic encoding routines for numerical features by simple standardization and normalization
    """
    min_val_pre, max_val_pre = data[numerical_features].min().to_dict(), data[numerical_features].max().to_dict()
    scaler = StandardScaler() if standardize else None
    if scaler is not None:
        data[numerical_features] = scaler.fit_transform(data[numerical_features])
    
    scaler = MinMaxScaler() if normalize else None
    if scaler is not None:
        data[numerical_features] = scaler.fit_transform(data[numerical_features])
    
    min_val, max_val = data[numerical_features].min().to_dict(), data[numerical_features].max().to_dict()
    
    data_encoding_log = {'standardize': standardize, 'normalize': normalize, 'info': {
        feature: {
            'min_pre': min_val_pre[feature],
            'max_pre': max_val_pre[feature],
            'min_post': min_val[feature],
            'max_post': max_val[feature]
        } for feature in numerical_features
    }}
    
    return data, data_encoding_log


def basic_categorical_encoding(
    data: pd.DataFrame,
    categorical_features: List[str],
    max_categories: int = 10
) -> Tuple[pd.DataFrame, dict]:
    """
    Basic encoding routines for categorical features by one-hot encoding with a max number of 10 categories
    """
    
    dim_pre = len(data.columns)
    encoder = OneHotEncoder(
        sparse_output=False, handle_unknown='ignore', max_categories=max_categories, drop = 'first'
    )
    encoded_features = encoder.fit_transform(data[categorical_features])
    encoded_feature_names = encoder.get_feature_names_out(categorical_features)
    encoded_df = pd.DataFrame(encoded_features, columns=encoded_feature_names, index=data.index)
    data = pd.concat([data.drop(columns=categorical_features), encoded_df], axis=1)

    # encoding log
    data_encoding_log = {
        'max_categories': max_categories,
        'cat_dim_pre': len(categorical_features),
        'cat_dim_post': len(encoded_feature_names),
        'dim_pre': dim_pre,
        'dim_post': len(data.columns),
        'encoded_cat_features': encoded_feature_names.tolist()
    }
    
    return data, data_encoding_log
